from typing import Union

from .module import BaseBus, BusFilter, BusInject
from .event import Event


class EventBus(BaseBus, BusFilter, BusInject):
    """
    事件总线
    """

    def __init__(self, max_concurrent_tasks: int = 10):
        BaseBus.__init__(self, max_concurrent_tasks)
        BusFilter.__init__(self)
        BusInject.__init__(self)

    async def publish(self, event: Union[Event, str], *args, **kwargs) -> None:
        await BusInject.resolve(self, event, args, kwargs)
        if await BusFilter.resolve(self, event, args, kwargs):
            return
        await super().publish(event, *args, **kwargs)

    def clear(self):
        super().clear()



from .event import Event
from .event_callback import EventCallback
from .sync_event_callback import SyncEventCallback
from .async_event_callback import AsyncEventCallback
from .event_callback_factory import EventCallbackFactory
from .event_callback_container import EventCallbackContainer



from abc import ABC, abstractmethod
from typing import Union

from ..event import Event


class BaseModule(ABC):
    """
    EventBus 模块基类
    """

    @abstractmethod
    async def resolve(self, event: Union[Event, str], args, kwargs) -> bool:
        raise NotImplementedError

    @abstractmethod
    def clear(self) -> None:
        raise NotImplementedError



from typing import Any

from .event_callback import EventCallback, T


class SyncEventCallback(EventCallback):
    def __init__(self, callback: T, weight: int = 1):
        super().__init__(callback, weight)
        self._async = False

    def __call__(self, *args, **kwargs) -> Any:
        return self.callback(*args, **kwargs)




from typing import Awaitable, Callable, Optional, Type, Union

from loguru import logger

from .base_module import BaseModule
from ..event import Event, EventCallbackContainer

FilterCallback: Type = Callable[..., Union[bool, Awaitable[bool]]]


class BusFilter(BaseModule):
    """
    事件总线过滤器模块, 负责对事件进行过滤, 同时判断是否继续传播该事件
    """

    def __init__(self):
        self._filters: dict[str, EventCallbackContainer] = {}
        self._global_filters: EventCallbackContainer = EventCallbackContainer()

    def clear(self):
        self._filters.clear()
        self._global_filters.clear()

    async def resolve(self, event: Union[Event, str], args, kwargs) -> bool:
        if await self._apply_global_filter(event, *args, **kwargs):
            return True
        if await self._apply_filter(event, *args, **kwargs):
            return True
        return False

    async def _apply_filter(self, event: Union[Event, str], *args, **kwargs) -> bool:
        if event in self._filters:
            for callback in self._filters[event].sync_callback:
                if callback(*args, **kwargs):
                    return True
            for callback in self._filters[event].async_callback:
                if await callback(*args, **kwargs):
                    return True
        return False

    async def _apply_global_filter(self, event: Union[Event, str], *args, **kwargs) -> bool:
        for callback in self._global_filters.sync_callback:
            if callback(event, *args, **kwargs):
                return True
        for callback in self._global_filters.async_callback:
            if await callback(event, *args, **kwargs):
                return True
        return False

    def global_event_filter(self, weight: int = 1) -> Callable[[FilterCallback], FilterCallback]:
        """
        全局过滤器修饰器\n
        使用修饰器来向事件总线注册一个全局事件过滤器函数，这个过滤器函数可以是异步函数，也可以是同步函数。\n
        传递函数的时候可以同时传递权重，权重大的过滤器会被优先调用\n
        过滤器返回值:\n
        如果返回True，则代表此事件被截断，不再向下传播。\n
        如果返回False，则此事件继续传播\n
        示例::\n
            @event_bus.on_global_event_filter()
            def message_logger(message, *_, **__):
                print(message)
            # 异步函数也可以
            @event_bus.on_global_event_filter()
            async def message_recoder(message, *_, **__):
                await ...
            # 可以同时传递权重
            @event_bus.on_global_event_filter(10)
            async def message_filter(message, *_, **__):
                await ...

        :param weight: 事件的选择权重
        """

        def decorator(func: FilterCallback):
            self.add_global_filter(func, weight)
            return func

        return decorator

    def add_global_filter(self, callback: FilterCallback, weight: int = 1) -> None:
        """
        注册全局过滤器\n
        用于注册全局过滤器的函数，也可以单独使用\n
        示例::\n
            def message_logger(message, *_, **__):
                print(message)
            async def message_recoder(message, *_, **__):
                await ...
            async def message_filter(message, *_, **__):
                await ...
            event_bus.add_global_filter(message_logger)
            event_bus.add_global_filter(message_recoder)
            event_bus.add_global_filter(message_filter, 10)

        :param callback: 事件回调函数
        :param weight: 事件的选择权重
        """
        self._global_filters.add_callback(callback, weight)
        logger.debug(f"Global filter {callback.__name__} has been added, weight={weight}")

    def remove_global_filter(self, callback: FilterCallback) -> None:
        """
        移除全局过滤器\n
        示例::\n
            def message_logger(message, *_, **__):
                print(message)
            event_bus.remove_global_filter(message_logger)

        :param callback: 事件回调函数
        """
        self._global_filters.remove_callback(callback)

    def event_filter(self, event: Union[Event, str], weight: int = 1) -> Callable[[FilterCallback], FilterCallback]:
        """
        事件过滤器修饰器\n
        使用修饰器来向事件总线注册一个事件过滤器函数，这个过滤器函数可以是异步函数，也可以是同步函数。\n
        事件类型可以是字符串，表示自定义类型，也可以写一个继承Event类的枚举类来管理事件。\n
        示例::\n
            @event_bus.event_filter('message_create')
            def message_logger(message, *_, **__):
                print(message)
            # 异步函数也可以
            @event_bus.event_filter('message_create')
            async def message_recoder(message, *_, **__):
                await ...
            # 可以同时传递权重
            @event_bus.event_filter('message_create', 10)
            async def message_filter(message, *_, **__):
                await ...

        :param event: 要订阅的事件
        :param weight: 事件的选择权重
        """

        def decorator(func: FilterCallback):
            self.add_filter(event, func, weight)
            return func

        return decorator

    def add_filter(self, event: Union[Event, str], callback: FilterCallback, weight: int = 1) -> None:
        """
        注册事件过滤器\n
        用于注册事件过滤器的函数，也可以单独使用\n
        示例::\n
            def message_logger(message, *_, **__):
                print(message)
            async def message_recoder(message, *_, **__):
                await ...
            async def message_filter(message, *_, **__):
                await ...
            event_bus.add_filter('message_create', message_logger)
            event_bus.add_filter('message_create', message_recoder)
            event_bus.add_filter('message_create', message_filter, 10)

        :param event: 要订阅的事件
        :param callback: 事件回调函数
        :param weight: 事件的选择权重
        """
        if event not in self._filters:
            self._filters[event] = EventCallbackContainer()
        self._filters[event].add_callback(callback, weight)
        logger.debug(f"Event filter {callback.__name__} has been added to event {event}, weight={weight}")

    def remove_filter(self, event: Union[Event, str], callback: FilterCallback) -> None:
        """
        移除事件过滤器

        示例::

            def message_logger(message, *_, **__):
                print(message)
            event_bus.remove_filter('message_create', message_logger)

        :param event: 要订阅的事件
        :param callback: 事件回调函数
        """
        if event in self._filters:
            self._filters[event].remove_callback(callback)



from asyncio import Semaphore, gather, run
from abc import ABC, abstractmethod
from typing import Any, Type, Union, Callable, Awaitable

from loguru import logger

from .module_exceptions import MultipleError
from ..event import Event, EventCallbackContainer

SubScriberCallback: Type = Callable[..., Union[Any, Awaitable[Any]]]


class BaseBus(ABC):
    """
    EventBus 事件基类，提供最基础的事件订阅和触发服务。
    :param max_concurrent_tasks: 异步任务的最大任务数
    """

    def __init__(self, max_concurrent_tasks: int = 10):
        self._subscribers: dict[str, EventCallbackContainer] = {}
        self._semaphore = Semaphore(max_concurrent_tasks)
        self._raise_exception = False

    def on(self, event: Union[Event, str], *, weight: int = 1) -> Callable:
        """
        订阅事件修饰器\n
        使用修饰器来向事件总线注册一个事件处理函数，这个事件处理函数可以是异步函数，也可以是同步函数。\n
        事件类型可以是字符串，表示自定义类型，也可以写一个继承Event类的枚举类来管理事件。\n
        示例::\n
            @event_bus.on('message_create')
            def message_logger(message, *_, **__):
                print(message)
            # 异步函数也可以
            @event_bus.on('message_create')
            async def message_recoder(message, *_, **__):
                await ...

        :param event: 要订阅的事件
        :param weight: 事件的选择权重
        :return: 实际上的修饰器
        """

        def decorator(func: SubScriberCallback):
            self.subscribe(event, func, weight=weight)
            logger.debug(f"{func.__name__} has subscribed to {event}, weight={weight}")
            return func

        return decorator

    def subscribe(self, event: Union[Event, str], callback: SubScriberCallback, *, weight: int = 1) -> None:
        """
        订阅事件\n
        用于订阅事件修饰器内部的函数，也可以单独使用\n
        详细信息请查看\n
        示例::\n
            def message_logger(message, *_, **__):
                print(message)
            async def message_recoder(message, *_, **__):
                await ...
            event_bus.subscribe('message_create', message_logger)
            event_bus.subscribe('message_create', message_recoder)

        :param event: 要订阅的事件
        :param callback: 事件回调函数
        :param weight: 事件的选择权重
        """
        if event not in self._subscribers:
            self._subscribers[event] = EventCallbackContainer()
        self._subscribers[event].add_callback(callback, weight)

    def unsubscribe(self, event: Union[Event, str], callback: SubScriberCallback) -> None:
        """
        取消订阅事件\n
        示例::\n
            def message_logger(message, *_, **__):
                print(message)
            event_bus.unsubscribe('message_create', message_logger)

        :param event: 要订阅的事件
        :param callback: 事件回调函数
        """
        if event in self._subscribers:
            self._subscribers[event].remove_callback(callback)

    def publish_sync(self, event: Union[Event, str], *args, **kwargs) -> None:
        """
        以同步方式触发事件（底层还是异步执行）\n
        示例:
            publish_sync('message_create', "This is a message", user="Half")

        :param event: 要触发的事件
        """
        run(self.publish(event, *args, **kwargs))

    async def _run_with_semaphore(self, coroutine: Callable, *args, **kwargs):
        """
        带有限制器的异步函数执行器
        :param coroutine: 原异步函数
        """
        async with self._semaphore:
            return await coroutine(*args, **kwargs)

    @abstractmethod
    async def publish(self, event: Union[Event, str], *args, **kwargs) -> None:
        """
        异步触发事件\n
        执行顺序:
            函数的权重越大，越先被执行，同步函数优先于异步函数执行。\n
            其中，最后的异步事件处理函数权重没有作用，因为会使用asyncio.gather并发执行，执行先后顺序也就失去了意义\n
        示例:
            await publish('message_create', "This is a message", user="Half")

        :param event: 要触发的事件
        """
        if event in self._subscribers:
            exceptions = []

            for callback in self._subscribers[event].sync_callback:
                try:
                    callback(*args, **kwargs)
                except Exception as e:
                    if self._raise_exception:
                        raise e
                    exceptions.append(e)

            async_handlers = [self._run_with_semaphore(callback, *args, **kwargs) for callback in
                              self._subscribers[event].async_callback]

            if self._raise_exception:
                await gather(*async_handlers, return_exceptions=False)
            else:
                results = await gather(*async_handlers, return_exceptions=True)
                if not (len(results) == 1 and results[0] is None):
                    exceptions.extend(results)

            if (exception_size := len(exceptions)) != 0:
                if exception_size == 1:
                    raise exceptions[0]
                else:
                    raise MultipleError(exceptions)

    @abstractmethod
    def clear(self):
        self._subscribers.clear()

    @property
    def raise_exception_immediately(self) -> bool:
        return self._raise_exception

    @raise_exception_immediately.setter
    def raise_exception_immediately(self, value: bool) -> None:
        self._raise_exception = value
